# -*- coding: utf-8 -*-

import datetime

from uuid import uuid4
from re import compile as recompile

from django.contrib.auth import authenticate
from django.contrib.auth.models import User
from django.db import IntegrityError
from django.db.models import Count
from django.http import HttpResponse, HttpResponseNotAllowed, HttpResponseForbidden, HttpResponseBadRequest
from django.shortcuts import get_object_or_404

from scrobbles.models import Artist, ArtistName, Track, TrackName, Scrobble


def get_tracks_count(user, dfrom, dto):
    return Track.objects.filter(scrobble__user=user, scrobble__date__range=(dfrom, dto)
                            ).annotate(count=Count('scrobble'))

def get_arts_count(user, dfrom, dto, limit=None, name=None):
    request = Artist.objects.filter(
                    track__scrobble__user=user, track__scrobble__date__range=(dfrom, dto)
            ).annotate(
                    count=Count('track__scrobble__track__artist')).order_by('-count')
    if name:
        request = request.filter(artistname__name__icontains=name)
    if limit:
        return request[:limit]
    return request

def get_arts_scrobbles(user, dfrom, dto, uuid):
    artist = get_object_or_404(Artist, scid=uuid)
    return Scrobble.objects.filter(
                    track__artist=artist,
                    user=user, date__range=(dfrom, dto)
            )

def last_year_as_epoch():
    year_ago = datetime.datetime.utcnow() - datetime.timedelta(days=365)
    return year_ago.strftime('%s')

def get_track_stats(user=None, day=False, week=False, month=False):
    if user == None:
        return None
    dfrom = datetime.datetime.now(tz=utc) - datetime.timedelta(days=365)
    limit = 3
    dto = datetime.datetime.now(tz=utc) + datetime.timedelta(days=7)
    tracks = get_tracks_count(user, dfrom, dto)
    return [{'name':trk.name,
             'count':trk.count,
             'artist':trk.artist.name  }
             for trk in tracks.order_by('-count')
             if trk.count >= limit
           ]

def get_art_stats(user=None, day=False, week=False, month=False):
    if user == None:
        return None
    if day == True:
        dfrom = datetime.datetime.now(tz=utc) - datetime.timedelta(hours=24)
        limit = 1
    if week == True:
        dfrom = datetime.datetime.now(tz=utc) - datetime.timedelta(days=7)
        limit = 4
    if month == True:
        dfrom = datetime.datetime.now(tz=utc) - datetime.timedelta(days=31)
        limit = 8
    if not dfrom:
        return None
    dto = datetime.datetime.now(tz=utc) + datetime.timedelta(days=7)
    arts = get_arts_count(user, dfrom, dto)
    return [{'name':art.name,
               'count':art.count,}
             for art in arts if art.count >= limit
           ]


ZERO = datetime.timedelta(0)

# A UTC class.
class UTC(datetime.tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO

utc = UTC()

class UUID(object):

    def __new__(self, value):
        pattern = r'[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}'
        regexp = recompile(pattern)
        if not regexp.match(value.lower()):
            raise ValueError(u'Malformed MusicBrainz ID: %s' % value)


class ScrobbleObject(object):
    """
    A scrobble must provide at least:
        * a timestamp as unix time
        * an artist name (UTF-8 encoded?)
        * a title name (UTF-8 encoded?)

    Everything else is optional.
    Musicbrainz ID are usefull, especially for artists, but not enough used out
    there to enforce them…
    """
    defaults = {'timestamp': None,  # mandatory
                'art': None,        # mandatory
                'tit': None,        # mandatory
                'alb': None,        # optional
                'art_mbid': None,   # optional but kinda helpful ;)
                'tit_mbid': None,   # optional but kinda helpful ;)
                'alb_mbid': None,   # optional
                }

    def __init__(self, data):
        self.request = data  # Special QueryDict dictionary, cf. file:///usr/share/doc/python-django-doc/html/ref/request-response.html#querydict-objects
        self.__dict__.update(ScrobbleObject.defaults)
        self._get_querydict()

    def _get_querydict(self):
        for k, v in self.request.iteritems():
            if k == 'timestamp':
                v = datetime.datetime.fromtimestamp(int(v), tz=utc)
            self.__dict__.update({k:v})

    def _check_provided_values(self):
        """Check either the scrobble is complete or not"""
        return (self.timestamp and self.art and self.tit)

    def _check_timestamp(self):
        """Controls timestamp"""
        return True

    def check_scrobble(self):
        return self._check_provided_values()


class APIObject(object):

    def __call__(self, request, username, **kwargs):
        self.__dict__.update(**kwargs)
        self.request = request
        # Look up the user and throw a 404 if it doesn't exist
        self.user = get_object_or_404(User, username=username)
        # Try to locate a handler method.
        try:
            callback = getattr(self, "do_%s" % request.method)
        except AttributeError:
            # This class doesn't implement this HTTP method, so return
            # a 405 ("Method Not Allowed") response and list the
            #allowed methods.
            allowed_methods = [m.lstrip("do_") for m in dir(self) if m.startswith("do_")]
            return HttpResponseNotAllowed(allowed_methods)
        # Check and store HTTP basic authentication, even for methods that
        # don't require authorization.
        self.authenticate()
        #if not self.authenticate():
        #    return self.unauthorized()
        valid, mess = self.validate()
        if not valid:
            return self.bad_request(content=mess)
        # Call the looked-up method
        return callback()

    def authenticate(self):
        self.authenticated_user = None
        if 'HTTP_AUTHORIZATION' in self.request.META:
            authmeth, auth = self.request.META['HTTP_AUTHORIZATION'].split(' ', 1)
            if authmeth.lower() == 'basic':
                auth = auth.strip().decode('base64')
                username, password = auth.split(':', 1)
                self.authenticated_user = authenticate(username=username, password=password)
                return True
        return False

    def unauthorized(self, content=None):
        response = HttpResponse(content=content)
        response.status_code = 401
        response['WWW-Authenticate'] = 'Basic realm="Scrobbler server"'
        return response

    def forbidden(self):
        response = HttpResponseForbidden()
        response["WWW-Authenticate"] = 'Basic realm="Scrobbler server"'
        return response

    def bad_request(self, content=None):
        if not content:
            content = u'Malformed request!'
        return HttpResponseBadRequest(content=content)

    def ok(self):
        response = HttpResponse()
        response.status_code = 200
        return response

    def get_filters(self, def_from='0', def_to='2000000000'):
        """
        Default upper limit to year 2033 == 2000000000s
        Default lower limit is epoch origin
        """
        epoch_from = self.request.GET.get('from', def_from)
        epoch_to   = self.request.GET.get('to', def_to)
        limit      = self.request.GET.get('limit', None)
        name       = self.request.GET.get('name', None)
        if limit:
            limit = int(limit)
        dfrom = datetime.datetime.fromtimestamp(int(epoch_from), tz=utc)
        dto   = datetime.datetime.fromtimestamp(int(epoch_to), tz=utc)
        return dfrom, dto, limit, name

    def validate(self):
        """
        Returns (Bool, message)
        If Boolean in False then message is not none and should contains info
        """
        return True, None


def update_obj_as_mbid(obj, mbid):
    """
    Updates object's scid except if it's already affected to another entry.
    TODO: implement a merging method in this case.
    """
    obj.scid = mbid
    obj.is_mbid = True
    try:
        obj.save()
    except IntegrityError:
        pass
    return obj

def set_art(scrobble):
    """"""
    name = scrobble.art
    mbid = scrobble.art_mbid
    # Looks for mbid entry
    if mbid:
        try:
            return Artist.objects.get(scid=mbid)
        except Artist.DoesNotExist:
            # log about scrobble with mbid provided
            pass
    # Now looks for entry by name.
    query = ArtistName.objects.select_related().filter(name=name)
    if len(query) > 1:
        #  Same artist name but different scid
        raise Exception('Multiple artist name: %s' % name)
    elif len(query) == 1:
        return query[0].artist
    # Creates new Artist / ArtistName entries
    if not mbid:
        mbid = str(uuid4())
    art = Artist.objects.create(scid=mbid)
    ArtistName.objects.create(name=name, artist=art)
    #TODO: log about created artist
    return art

def set_trk(scrobble, artist):
    name = scrobble.tit
    mbid = scrobble.tit_mbid
    if mbid:
        try:
            return Track.objects.get(scid=mbid)
        except Track.DoesNotExist:
            pass
    try:
        track_name = TrackName.objects.get(name=name, track__artist=artist)
        return track_name.track
    except TrackName.DoesNotExist:
        pass
    #except MultipleObjectsReturned: # Exception from django.core.exceptions
        #pass ## Should not occur!!!
    # Creating New Track / TrackName entries
    if not mbid:
        mbid = str(uuid4())
    track = Track.objects.create(artist=artist, scid=mbid)
    TrackName.objects.create(name=name, track=track)
    return track

def add_scrobble(scrobble, user):
    """
    At this point the API should have checked presence of mandatory fileds.
    data is potentially:
                >>> {'timestamp': int,  # mandatory
                >>>  'art': str,        # mandatory
                >>>  'tit': str,        # mandatory
                >>>  'alb': str,        # optional
                >>>  'art_mbid': uuid,  # optional but kinda helpful ;)
                >>>  'tit_mbid': uuid,  # optional but kinda helpful ;)
                >>>  'alb_mbid': uuid,  # optional
    """
    artist = set_art(scrobble)
    track = set_trk(scrobble, artist)
    #if scrobble.alb or scrobble.alb_mbid:
    #    set_album(scrobble)
    scrobble = Scrobble.objects.create(user=user, track=track, date=scrobble.timestamp)

# vim: ai ts=4 sw=4 sts=4 expandtab
