#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Doc:
#
# URLEncoding for queries (ie. + encoded space for instance)
#   https://tools.ietf.org/html/rfc1866#section-8.2
#

import urllib.request
import urllib.parse

from urllib.error import URLError, HTTPError
from http.client import BadStatusLine

from time import time
from datetime import datetime, timedelta


WAIT_BETWEEN_REQUESTS = timedelta(0, 0.9)

class ScrobbleError(Exception):
    """Generic Error"""
    pass

class ScrobbleHTTPError(ScrobbleError):
    """HTTP scrobbler Error"""

    def __init__(self, expression):
        self.http_out = None
        self.expression = expression
        self.code = expression.code
        if hasattr(expression, 'read'):
            self.http_out = expression.read().decode('utf-8')

    def __str__(self):
        msg = 'HTTPError {0}: {1}'.format(self.code, self.expression.reason)
        return msg

class ScrobbleUnhandledError(ScrobbleError):
    pass

class Throttle(object):
    def __init__(self, wait):
        self.wait = wait
        self.last_called = datetime.now()

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            while self.last_called + self.wait > datetime.now():
                sleep(0.2)
            result = func(*args, **kwargs)
            self.last_called = datetime.now()
            return result
        return wrapper


class Track(object):
    """
    Here is how to instanciate a Track object:

        >>> track = Track(**{'artist': 'The Tool', 'title': 'Opiate', 'musicbrainz_artistid': '1234-345-45'})

        Track ojects are just to set default values for track retrived from audio player (cf. Track.default dictionary).
    """
    default = {
            'title'                 : '',
            'artist'                : '',
            'album'                 : '',
            'albumartist'           : '',
            'musicbrainz_trackid'   : None,
            'musicbrainz_artistid'  : None,
            'musicbrainz_albumid'   : None,
            'timestamp'             : 1,
             }

    def __init__(self, **kwargs):
        self.__dict__.update(**self.__class__.default)
        self.__dict__.update(**kwargs)

def get_api_options(track):
    mapping = {'title':  'tit',
               'artist': 'art',
               'album':  'alb',
               'musicbrainz_trackid':  'tit_mbid',
               'musicbrainz_artistid': 'art_mbid',
               'musicbrainz_albumid':  'alb_mbid',
               'timestamp'          :  'timestamp',
            }
    options = {}
    for k, v in mapping.items():
        if track.__getattribute__(k):
            options[v] = track.__getattribute__(k)
    return options

@Throttle(WAIT_BETWEEN_REQUESTS)
def scrobble(trk=None, user=None, pwd=None, srv=None):
    sc_url = '{server}/api/{user}/scrobbles/'
    url = sc_url.format(user=user, server=srv)
    params = urllib.parse.urlencode(get_api_options(trk))
    auth_handler = urllib.request.HTTPBasicAuthHandler()
    auth_handler.add_password(realm='Scrobbler server',
                              uri=url,
                              user=user,
                              passwd=pwd)
    opener = urllib.request.build_opener(auth_handler)
    urllib.request.install_opener(opener)
    params = params.encode('utf-8')
    try:
        urllib.request.urlopen(url, params)
    except HTTPError as err:
        raise ScrobbleHTTPError(err)
    except URLError as err:
        raise ScrobbleError(err)
    except BadStatusLine as err:
        raise ScrobbleError(err)
    except Exception as err:
        raise ScrobbleUnhandledError(err)

def main():
    parser = argparse.ArgumentParser(prog='scrobbler client')
    for opt in ARGPARSE:  # Populates ArgumentParser
        parser.add_argument(*opt[0], **opt[1])
    args = parser.parse_args()
    user = vars(args).get('user')
    track = Track(**vars(args))  # Shipping 'user' key/val as well, but remains unused in get_api_options.
    try:
        scrobble(trk=track, user=user, srv=SERVER, pwd=PWD)
    except Exception as err:
        print(err)


# Script starts here
if __name__ == '__main__':
    ############################
    # CONFIG
    SERVER = 'http://127.0.0.1/scserver'
    USER   = 'myusername'
    PWD    = 'myS3cr34'
    #
    ############################
    try:
        from config import SERVER, USER, PWD
    except ImportError:
        pass
    ARGPARSE = [
                [['artist'], {'help':'track\'s artist'},],
                [['title'],  {'help':'track\'s title'},],
                [['--user'], {'help':'username to know which account to scrobble to', 'default':USER}],
                [['--arid'], {'dest':'musicbrainz_artistid'},],
                [['--trid'], {'dest':'musicbrainz_trackid'},],
                [['--timestamp'], {'dest':'timestamp', 'default':int(time()),
                    'help':'timestamp for the scrobble, time in second since the '+
                    'epoch (as an integer). If not provided, defaults to the current time.'}],
                ]
    import argparse
    main()

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab

